@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: hald_dbus.c,v 1.13 2004/09/19 16:32:15 david Exp $
+ * CVSID: $Id: hald_dbus.c,v 1.16.2.1 2004/12/09 02:35:02 david Exp $
  *
  * dbus.c : D-BUS interface of HAL daemon
  *
@@ -30,6 +30,8 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <unistd.h>
+#include <sys/types.h>
 
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib-lowlevel.h>
@@ -910,6 +912,37 @@ device_get_property_type (DBusConnection * connection,
 }
 
 
+static dbus_bool_t 
+sender_has_privileges (DBusConnection *connection, DBusMessage *message)
+{
+	DBusError error;
+	unsigned long user_uid;
+	const char *user_base_svc;
+
+	user_base_svc = dbus_message_get_sender (message);
+	if (user_base_svc == NULL) {
+		HAL_WARNING (("Cannot determine base service of caller"));
+		return FALSE;
+	}
+
+	HAL_DEBUG (("base_svc = %s", user_base_svc));
+
+	dbus_error_init (&error);
+	if ((user_uid = dbus_bus_get_unix_user (connection, user_base_svc, &error)) 
+       == (unsigned long) -1) {
+		HAL_WARNING (("Could not get uid for connection"));
+		return FALSE;
+	}
+
+	HAL_INFO (("uid for caller is %ld", user_uid));
+
+	if (user_uid != 0 && user_uid != geteuid()) {
+		HAL_WARNING (("uid %d is doesn't have the right priviledges", user_uid));
+		return FALSE;
+	}
+
+	return TRUE;
+}
 
 /** Set a property on a device.
  *
@@ -952,6 +985,11 @@ device_set_property (DBusConnection * connection, DBusMessage * message)
 	}
 	key = dbus_message_iter_get_string (&iter);
 
+	if (!sender_has_privileges (connection, message)) {
+		raise_permission_denied (connection, message, "SetProperty: not privileged");
+		return DBUS_HANDLER_RESULT_HANDLED;
+	}
+
 	HAL_DEBUG (("udi=%s, key=%s", udi, key));
 
 	device = hal_device_store_find (hald_get_gdl (), udi);
@@ -1046,6 +1084,7 @@ device_set_property (DBusConnection * connection, DBusMessage * message)
  *  void Device.AddCapability(string capability)
  *
  *    raises org.freedesktop.Hal.NoSuchDevice, 
+ *    raises org.freedesktop.Hal.PermissionDenied, 
  *  </pre>
  *
  *  @param  connection          D-BUS connection
@@ -1065,6 +1104,11 @@ device_add_capability (DBusConnection * connection, DBusMessage * message)
 
 	HAL_TRACE (("entering"));
 
+	if (!sender_has_privileges (connection, message)) {
+		raise_permission_denied (connection, message, "AddCapability: not privileged");
+		return DBUS_HANDLER_RESULT_HANDLED;
+	}
+
 	udi = dbus_message_get_path (message);
 
 	d = hal_device_store_find (hald_get_gdl (), udi);
@@ -1120,6 +1164,7 @@ device_add_capability (DBusConnection * connection, DBusMessage * message)
  *
  *    raises org.freedesktop.Hal.NoSuchDevice, 
  *           org.freedesktop.Hal.NoSuchProperty
+ *           org.freedesktop.Hal.PermissionDenied
  *  </pre>
  *
  *  @param  connection          D-BUS connection
@@ -1139,6 +1184,11 @@ device_remove_property (DBusConnection * connection, DBusMessage * message)
 
 	udi = dbus_message_get_path (message);
 
+	if (!sender_has_privileges (connection, message)) {
+		raise_permission_denied (connection, message, "RemoveProperty: not privileged");
+		return DBUS_HANDLER_RESULT_HANDLED;
+	}
+
 	d = hal_device_store_find (hald_get_gdl (), udi);
 	if (d == NULL)
 		d = hal_device_store_find (hald_get_tdl (), udi);
@@ -1289,7 +1339,7 @@ device_query_capability (DBusConnection * connection,
 		char **capsv, **iter;
 
 		capsv = g_strsplit (caps, " ", 0);
-		for (iter = capsv; iter != NULL; iter++) {
+		for (iter = capsv; *iter != NULL; iter++) {
 			if (strcmp (*iter, capability) == 0) {
 				rc = TRUE;
 				break;
