@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: block_class_device.c,v 1.74 2004/09/20 09:00:14 david Exp $
+ * CVSID: $Id: block_class_device.c,v 1.86.2.8 2005/01/21 15:46:55 david Exp $
  *
  * Block device class
  *
@@ -71,8 +71,6 @@
  * @{
  */
 
-static void etc_mtab_process_all_block_devices (dbus_bool_t force);
-
 static char *block_class_compute_udi (HalDevice * d, int append_num);
 
 #if 0
@@ -157,18 +155,36 @@ ClassDeviceHandler block_class_handler;
 
 static dbus_bool_t
 block_class_accept (ClassDeviceHandler *self,
-		    const char *path,
+		    const char *sysfs_path,
 		    struct sysfs_class_device *class_device)
 {
+	int instance;
+	HalDevice *d;
 
 	/*HAL_INFO (("path = %s, classname = %s", 
-	  path, self->sysfs_class_name));*/
+	  sysfs_path, self->sysfs_class_name));*/
+
+	/* skip legacy floppies for now, until we get proper sysfs links to the
+	   platform device and switch over to merge floppies into that device.
+	   Legacy floppies handled until that in platform_bus_device.c */
+	if (sscanf (class_device->name, "fd%d", &instance) == 1)
+		return FALSE;
+
+	/* ignore block devices we already have added (see where we refuse
+	 * to remove them in block_class_removed)
+	 */
+	d = hal_device_store_match_key_value_string (hald_get_gdl (), "linux.sysfs_path", sysfs_path);
+	if (d != NULL) {
+		HAL_INFO (("Ignoring block device add for %s since it is already in HAL (ide-cs?)", sysfs_path));
+		return FALSE;
+	}
 
 	/* only care about given sysfs class name */
 	if (strcmp (class_device->classname, self->sysfs_class_name) == 0) {
 		return TRUE;
 	}
 
+
 	return FALSE;
 }
 
@@ -180,7 +196,6 @@ block_class_visit (ClassDeviceHandler *self,
 {
 	HalDevice *d;
 	HalDevice *parent;
-	char *parent_sysfs_path;
 	ClassAsyncData *cad;
 	struct sysfs_device *sysdevice;
 
@@ -202,10 +217,12 @@ block_class_visit (ClassDeviceHandler *self,
 	sysdevice = sysfs_get_classdev_device (class_device);
 
 	if (sysdevice == NULL) {
-		parent_sysfs_path = get_parent_sysfs_path (path);
+		parent = find_closest_ancestor (path);
 		hal_device_property_set_bool (d, "block.is_volume", TRUE);
 	} else {
-		parent_sysfs_path = sysdevice->path;
+		parent = hal_device_store_match_key_value_string (hald_get_gdl (), 
+								  "linux.sysfs_path_device", 
+								  sysdevice->path);
 		hal_device_property_set_bool (d, "block.is_volume", FALSE);
 	}
 
@@ -238,15 +255,21 @@ block_class_visit (ClassDeviceHandler *self,
 	 * sure to be added before us (we probe devices in the right order
 	 * and we reorder hotplug events)
 	 */
-	parent = hal_device_store_match_key_value_string (hald_get_gdl (), 
-							  "linux.sysfs_path_device", 
-							  parent_sysfs_path);
 	if (parent == NULL) {
 		hal_device_store_remove (hald_get_tdl (), d);
 		d = NULL;
 		goto out;
 	}
 
+	/* Never add e.g. /dev/hdd4 as child of /dev/hdd if /dev/hdd is without partitions (e.g. zip disks) */
+	if (hal_device_has_property (parent, "storage.no_partitions_hint")) {
+		if (hal_device_property_get_bool (parent, "storage.no_partitions_hint")) {
+			hal_device_store_remove (hald_get_tdl (), d);
+			d = NULL;
+			goto out;			
+		}
+	}
+
 	class_device_got_parent_device (hald_get_tdl (), parent, cad);
 
 out:
@@ -331,7 +354,7 @@ cdrom_get_properties (HalDevice *d, const char *device_file)
 	}
 	
 	/* while we're at it, check if we support media changed */
-	if (ioctl (fd, CDROM_MEDIA_CHANGED) >= 0) {
+	if (capabilities & CDC_MEDIA_CHANGED) {
 		hal_device_property_set_bool (d, "storage.cdrom.support_media_changed", TRUE);
 	} else {
 		hal_device_property_set_bool (d, "storage.cdrom.support_media_changed", FALSE);
@@ -358,6 +381,8 @@ cdrom_get_properties (HalDevice *d, const char *device_file)
 static void
 force_unmount (HalDevice * d)
 {
+	const char *storudi;
+	HalDevice *stordev;
 	const char *device_file;
 	const char *device_mount_point;
 	const char *umount_argv[4] = { "/bin/umount", "-l", NULL, NULL };
@@ -366,8 +391,19 @@ force_unmount (HalDevice * d)
 	int umount_exitcode;
 
 	device_file = hal_device_property_get_string (d, "block.device");
-	device_mount_point =
-	    hal_device_property_get_string (d, "volume.mount_point");
+	device_mount_point = hal_device_property_get_string (d, "volume.mount_point");
+
+	/* Only attempt to 'umount -l' if some hal policy piece are performing policy on the device */
+	storudi = hal_device_property_get_string (d, "block.storage_device");
+	if (storudi == NULL)
+		return;
+	stordev = hal_device_store_find (hald_get_gdl (), storudi);
+	if (stordev == NULL)
+		return;
+	HAL_INFO (("foo = %d", hal_device_property_get_bool (stordev, "storage.policy.should_mount")));
+	if ((!hal_device_has_property (stordev, "storage.policy.should_mount")) ||
+	    (!hal_device_property_get_bool (stordev, "storage.policy.should_mount")))
+		return;
 
 	umount_argv[2] = device_file;
 
@@ -485,6 +521,17 @@ volume_remove_from_gdl (HalDevice *device, gpointer user_data)
 	hal_device_store_remove (hald_get_gdl (), device);
 }
 
+static void
+volume_remove_from_gdl_sleep (HalDevice *device, gpointer user_data)
+{
+	g_signal_handlers_disconnect_by_func (device, volume_remove_from_gdl, 
+					      user_data);
+	/* extremely nasty hack */
+	HAL_INFO (("hnap 1"));
+	sleep (1);
+	hal_device_store_remove (hald_get_gdl (), device);
+}
+
 
 /** Check if a filesystem on a special device file is mounted
  *
@@ -571,38 +618,41 @@ volume_set_size (HalDevice *d, dbus_bool_t force)
 	storudi = hal_device_property_get_string (d, "block.storage_device");
 	stordev = hal_device_store_find (hald_get_gdl (), storudi);
 
+	sysfs_path = hal_device_property_get_string (d, "linux.sysfs_path");
+	/* no-partition volumes don't have a sysfs path */
+	if (sysfs_path == NULL)
+		sysfs_path = hal_device_property_get_string (stordev, "linux.sysfs_path");
+	
+	if (sysfs_path == NULL)
+		return;
+
+	snprintf (attr_path, SYSFS_PATH_MAX, "%s/size", sysfs_path);
+	attr = sysfs_open_attribute (attr_path);
+	if (sysfs_read_attribute (attr) >= 0) {
+		num_blocks = atoi (attr->value);
+		
+		hal_device_property_set_int (d, "volume.num_blocks", num_blocks);
+		hal_device_property_set_int (d, "volume.block_size", 512);
+		HAL_INFO (("volume.num_blocks = %d", num_blocks));
+		sysfs_close_attribute (attr);
+
+		/* for some devices we can't open the device file (IDE devices) so assume one block is 512 bytes */
+		hal_device_property_set_uint64 (d, "volume.size", ((dbus_uint64_t)(512))*((dbus_uint64_t)num_blocks));
+	}
+
 	if (force || hal_device_property_get_bool (stordev, "storage.media_check_enabled")) {
 
-		sysfs_path = hal_device_property_get_string (d, "linux.sysfs_path");
-		/* no-partition volumes don't have a sysfs path */
-		if (sysfs_path == NULL)
-			sysfs_path = hal_device_property_get_string (stordev, "linux.sysfs_path");
-
-		if (sysfs_path == NULL)
-			return;
-
 		device_file = hal_device_property_get_string (d, "block.device");
 
-		snprintf (attr_path, SYSFS_PATH_MAX, "%s/size", sysfs_path);
-		attr = sysfs_open_attribute (attr_path);
-		if (sysfs_read_attribute (attr) >= 0) {
-			num_blocks = atoi (attr->value);
-
-			hal_device_property_set_int (d, "volume.num_blocks", num_blocks);
-			HAL_INFO (("volume.num_blocks = %d", num_blocks));
-			sysfs_close_attribute (attr);
-		}
-
 		fd = open (device_file, O_RDONLY);
 		if (fd >= 0) {
 			if (ioctl (fd, BLKSSZGET, &block_size) == 0) {
 				hal_device_property_set_int (d, "volume.block_size", block_size);
 				HAL_INFO (("volume.block_size = %d", block_size));
 			}
-
 			if (ioctl (fd, BLKGETSIZE64, &vol_size) == 0) {
 				hal_device_property_set_uint64 (d, "volume.size", vol_size);
-				HAL_INFO (("volume.size = %lld", vol_size));
+				HAL_INFO (("volume.size = %llu", vol_size));
 			}
 
 			close (fd);
@@ -628,7 +678,14 @@ detect_disc (HalDevice *d, const char *device_file)
 	fd = open (device_file, O_RDONLY | O_NONBLOCK);
 	if (fd < 0)
 		return;
-	
+
+	/* Suggested by Alex Larsson to get rid of log spewage
+	 * on Alan's cd changer (RH bug 130649) */
+	if (ioctl (fd, CDROM_DRIVE_STATUS, CDSL_CURRENT) != CDS_DISC_OK) {
+		close (fd);
+		return;
+	}
+
 	/* check for audio/data/blank */
 	type = ioctl (fd, CDROM_DISC_STATUS, CDSL_CURRENT);
 	switch (type) {
@@ -822,7 +879,17 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 			break;
 			
 		case CDS_DISC_OK:
-			got_media = TRUE;
+			/* some CD-ROMs report CDS_DISK_OK even with an open
+			 * tray; if media check has the same value two times in
+			 * a row then this seems to be the case and we must not
+			 * report that there is a media in it. */
+			if (hal_device_property_get_bool (d, "storage.cdrom.support_media_changed") &&
+			    ioctl (fd, CDROM_MEDIA_CHANGED, CDSL_CURRENT) && 
+			    ioctl (fd, CDROM_MEDIA_CHANGED, CDSL_CURRENT)) {
+				/*HAL_INFO (("CD-ROM drive %s: media checking is broken, assuming no CD is inside.", device_file));*/
+			} else {
+			    got_media = TRUE;
+			}
 			break;
 			
 		case -1:
@@ -898,6 +965,8 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 	if (fd < 0)
 		return FALSE;
 
+	close (fd);
+
 	/* Now got_media==TRUE and fd>0.. So, Yay!, we got media
 	 *
 	 * See if we already got children (or children underway :-),
@@ -907,10 +976,9 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 
 	child = get_child_device_gdl (d);
 	if (child == NULL) {
-		get_child_device_tdl (d);
+		child = get_child_device_tdl (d);
 	}
 	if (child != NULL) {
-		close (fd);
 		return FALSE;
 	}
 
@@ -987,9 +1055,8 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 
 			HAL_INFO (("Detecting if %s contains a fs", device_file));
 
-			vid = volume_id_open_fd (fd);
+			vid = volume_id_open_node (device_file);
 			if (vid == NULL) {
-				close (fd);
 				g_object_unref (child);
 				return FALSE;
 			}
@@ -1001,12 +1068,12 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 					volume_id_close (vid);
 					vid = NULL;
 				} else {
-					close (fd);
 					g_object_unref (child);
 					volume_id_close (vid);
 					return FALSE;
 				}
 			}
+
 		}
 
 		/* Unfortunally, linux doesn't scan optical discs for partition
@@ -1019,7 +1086,6 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 		if (!no_partitions) {
 			if (vid != NULL)
 				volume_id_close (vid);
-			close (fd);
 			g_object_unref (child);
 			return FALSE;
 		}
@@ -1059,12 +1125,11 @@ detect_media (HalDevice * d, dbus_bool_t force_poll)
 			G_CALLBACK (class_device_move_from_tdl_to_gdl), cad);
 		hal_callout_device (child, TRUE);
 
-		volume_id_close (vid);
-		close (fd);
+		if (vid != NULL)
+			volume_id_close (vid);
 		return TRUE;
 	}
 
-	close (fd);
 	return FALSE;
 }
 
@@ -1105,6 +1170,7 @@ block_class_pre_process (ClassDeviceHandler *self,
 	const char *device_file;
 	dbus_bool_t has_removable_media = FALSE;
 	dbus_bool_t is_hotpluggable = FALSE;
+	dbus_bool_t requires_eject = FALSE;
 	char attr_path[SYSFS_PATH_MAX];
 	struct sysfs_attribute *attr;
 
@@ -1173,30 +1239,26 @@ block_class_pre_process (ClassDeviceHandler *self,
 				scsidev = d_it;
 				physdev = d_it;
 				physdev_udi = udi_it;
-				hal_device_property_set_string (
-					stordev, "storage.bus", "scsi");
+				hal_device_property_set_string (stordev, "storage.bus", "scsi");
 			}
 
 			if (strcmp (bus, "usb") == 0) {
 				physdev = d_it;
 				physdev_udi = udi_it;
 				is_hotpluggable = TRUE;
-				hal_device_property_set_string (
-					stordev, "storage.bus", "usb");
+				hal_device_property_set_string (stordev, "storage.bus", "usb");
 								
 				break;
 			} else if (strcmp (bus, "ieee1394") == 0) {
 				physdev = d_it;
 				physdev_udi = udi_it;
 				is_hotpluggable = TRUE;
-				hal_device_property_set_string (
-					stordev, "storage.bus", "ieee1394");
+				hal_device_property_set_string (stordev, "storage.bus", "ieee1394");
 				break;
 			} else if (strcmp (bus, "ide") == 0) {
 				physdev = d_it;
 				physdev_udi = udi_it;
-				hal_device_property_set_string (
-					stordev, "storage.bus", "ide");
+				hal_device_property_set_string (stordev, "storage.bus", "ide");
 				break;
 			}
 
@@ -1246,8 +1308,11 @@ block_class_pre_process (ClassDeviceHandler *self,
 		 * cause inifite loops of hotplug events, cf. broken ide-cs driver and
 		 * broken zip drives. Merely accessing the top-level block device if it
 		 * or any of it partitions are not mounted causes the loop.
+		 *
+		 * Also allow this for devices without removable media
 		 */
-		if (hal_device_property_get_bool (stordev, "storage.media_check_enabled")) {
+		if (hal_device_property_get_bool (stordev, "storage.media_check_enabled") ||
+		    !hal_device_property_get_bool (stordev, "storage.removable")) {
 			dbus_uint64_t size = 0;
 			const char *stordev_device_file;
 
@@ -1269,13 +1334,13 @@ block_class_pre_process (ClassDeviceHandler *self,
 				if (volume_id_probe(vid, VOLUME_ID_MSDOSPARTTABLE, 0, size) == 0) {
 					HAL_INFO (("Number of partitions = %d", vid->partition_count));
 
-					if (partition_number >= 0 && partition_number < vid->partition_count) {
+					if (partition_number > 0 && partition_number <= vid->partition_count) {
 						struct volume_id_partition *p;
-						p = &vid->partitions[partition_number];
+						p = &vid->partitions[partition_number-1];
 
 
 						hal_device_property_set_int (
-							d, "volume.partition.x86_type",
+							d, "volume.partition.msdos_part_table_type",
 							p->partition_type_raw);
 						
 						/* NOTE: We trust the type from the partition table
@@ -1304,13 +1369,13 @@ block_class_pre_process (ClassDeviceHandler *self,
 			}
 
 		} else {
-			/* gee, so at least set volume.fstype vfat,msdos,auto so 
-			 * mount(1) doesn't screw up and causes hotplug events
+			/* gee, so at least set volume.fstype to auto..
 			 *
 			 * GRRRR!!!
 			 */
-			hal_device_property_set_string (d, "volume.fstype", "vfat,auto");
-			hal_device_property_set_string (d, "volume.usage", "filesystem");
+			hal_device_property_set_string (d, "volume.fstype", "auto");
+			hal_device_property_set_string (d, "volume.fsusage", "filesystem");
+			volume_set_size (d, FALSE);
 		}
 		return;
 	}
@@ -1322,6 +1387,17 @@ block_class_pre_process (ClassDeviceHandler *self,
 	 ************************************************************/
 
 
+	snprintf (attr_path, SYSFS_PATH_MAX, "%s/removable", sysfs_path);
+	attr = sysfs_open_attribute (attr_path);
+	if (sysfs_read_attribute (attr) >= 0) {
+		if (attr->value [0] == '0')
+			has_removable_media = FALSE;
+		else
+			has_removable_media = TRUE;
+
+		sysfs_close_attribute (attr);
+	} 	
+
 	/* defaults */
 	hal_device_property_set_string (stordev, "storage.drive_type", "disk");
 
@@ -1333,43 +1409,19 @@ block_class_pre_process (ClassDeviceHandler *self,
 	HAL_INFO (("Bus type is %s!",
 		   hal_device_property_get_string (parent, "info.bus")));
 
-	if (strcmp (hal_device_property_get_string (parent, "info.bus"),
-			    "ide") == 0) {
+	if (strcmp (hal_device_property_get_string (parent, "info.bus"), "ide") == 0) {
 		const char *ide_name;
 		char *model;
 		char *media;
 
+		/* Be conservative and don't poll IDE drives at all */
+		hal_device_property_set_bool (d, "storage.media_check_enabled", FALSE);
 
-		/* blacklist the broken ide-cs driver */
-		if (physdev != NULL) {
-			size_t len;
-			char buf[256];
-			const char *physdev_sysfs_path;
-			
-			snprintf (buf, 256, "%s/devices/ide", sysfs_mount_path);
-			len = strlen (buf);
-			
-			physdev_sysfs_path = hal_device_property_get_string (physdev, "linux.sysfs_path");
-			
-			if (strncmp (physdev_sysfs_path, buf, len) == 0) {
-				hal_device_property_set_bool (stordev, "storage.media_check_enabled", FALSE);
-			}
-			
-			HAL_INFO (("Working around broken ide-cs driver for %s", physdev->udi));
-		}
-
-
-		ide_name = get_last_element (hal_device_property_get_string
-					     (d, "linux.sysfs_path"));
-
+		ide_name = get_last_element (hal_device_property_get_string (d, "linux.sysfs_path"));
 		model = read_single_line ("/proc/ide/%s/model", ide_name);
 		if (model != NULL) {
-			hal_device_property_set_string (stordev,
-							"storage.model",
-							model);
-			hal_device_property_set_string (d, 
-							"info.product",
-							model);
+			hal_device_property_set_string (stordev, "storage.model", model);
+			hal_device_property_set_string (d, "info.product", model);
 		}
 
 		/* According to the function proc_ide_read_media() in 
@@ -1378,38 +1430,21 @@ block_class_pre_process (ClassDeviceHandler *self,
 		 * "UNKNOWN"
 		 */
 		
-		/** @todo Given floppy how
-		 *        do we determine it's LS120?
-		 */
-			
-		media = read_single_line ("/proc/ide/%s/media",
-					  ide_name);
+		media = read_single_line ("/proc/ide/%s/media", ide_name);
 		if (media != NULL) {
-			hal_device_property_set_string (stordev, 
-							"storage.drive_type",
-							media);
+			hal_device_property_set_string (stordev, "storage.drive_type", media);
 			
 			/* Set for removable media */
 			if (strcmp (media, "disk") == 0) {
 				/* left blank */
 			} else if (strcmp (media, "cdrom") == 0) {
 				has_removable_media = TRUE;
+				/* cdroms are the only IDE devices that are safe to poll */
+				hal_device_property_set_bool (d, "storage.media_check_enabled", TRUE);
 			} else if (strcmp (media, "floppy") == 0) {
 				has_removable_media = TRUE;
-
-				/* I've got a LS120 that identifies as a
-				 * floppy; polling doesn't work so disable
-				 * media check and automount
-				 */
-				hal_device_property_set_bool (
-					d, "storage.media_check_enabled", FALSE);
-				hal_device_property_set_bool (
-					d, "storage.automount_enabled_hint", FALSE);
-
 			} else if (strcmp (media, "tape") == 0) {
 				has_removable_media = TRUE;
-
-				/* TODO: Someone test with tape drives! */
 			}			
 		}
 
@@ -1417,21 +1452,22 @@ block_class_pre_process (ClassDeviceHandler *self,
 		 * cause inifite loops of hotplug events, cf. broken ide-cs driver and
 		 * broken zip drives. Merely accessing the top-level block device if it
 		 * or any of it partitions are not mounted causes the loop.
+		 *
+		 * Also allow this when we don't have removable media.
 		 */
-		if (hal_device_property_get_bool (stordev, "storage.media_check_enabled")) {
+		if (hal_device_property_get_bool (stordev, "storage.media_check_enabled") ||
+		    !has_removable_media) {
 			const char *device_file;
 			struct drive_id *did;
 
 			device_file = hal_device_property_get_string (d, "block.device");
 			did = drive_id_open_node(device_file);
+			
 			if (drive_id_probe(did, DRIVE_ID_ATA) == 0) {
 				if (did->serial[0] != '\0')
-					hal_device_property_set_string (stordev, 
-									"storage.serial",
-									did->serial);
+					hal_device_property_set_string (stordev, "storage.serial", did->serial);
 				if (did->firmware[0] != '\0')
-					hal_device_property_set_string (stordev, 
-									"storage.firmware_version",
+					hal_device_property_set_string (stordev, "storage.firmware_version", 
 									did->firmware);
 			}
 			drive_id_close(did);
@@ -1444,8 +1480,6 @@ block_class_pre_process (ClassDeviceHandler *self,
 		const char *device_file;
 		struct drive_id *did;
 		const char *sysfs_path;
-		int scsi_host;
-		char *scsi_protocol;
 		
 		sysfs_path = hal_device_property_get_string (
 			d, "linux.sysfs_path");
@@ -1478,18 +1512,44 @@ block_class_pre_process (ClassDeviceHandler *self,
 		}
 
 		device_file = hal_device_property_get_string (d, "block.device");
-		did = drive_id_open_node(device_file);
-		if (drive_id_probe(did, DRIVE_ID_SCSI) == 0) {
-			if (did->serial[0] != '\0')
-				hal_device_property_set_string (stordev,
-								"storage.serial",
-								did->serial);
-			if (did->revision[0] != '\0')
-				hal_device_property_set_string (stordev, 
-								"storage.revision",
-								did->revision);
+
+		/* Only do drive_id on real SCSI devices - not on USB which uses emulated SCSI
+		 * since an INQUIRY on most USB devices may crash the storage device if the
+		 * transfer length isn't exactly 36 bytes.
+		 *
+		 * (See also Red Hat bug #145256)
+		 */
+		if (strcmp (hal_device_property_get_string (stordev, "storage.bus"), "scsi") == 0) {
+
+			did = drive_id_open_node(device_file);
+			if (drive_id_probe(did, DRIVE_ID_SCSI) == 0) {
+				if (did->serial[0] != '\0')
+					hal_device_property_set_string (stordev,
+									"storage.serial",
+									did->serial);
+				if (did->revision[0] != '\0')
+					hal_device_property_set_string (stordev, 
+									"storage.revision",
+									did->revision);
+			}
+			drive_id_close(did);
+		} 
+
+		/* see if this is really a SATA disk */
+#if 0
+		/* commented out until the value of ATA_IOC_GET_IO32 is available */
+		{
+			int fd;
+			unsigned char unused;
+
+			if ((fd = open (device_file, O_RDONLY|O_NDELAY)) != -1) {
+				if (ioctl (fd, ATA_IOC_GET_IO32, &unused) >= 0) {
+					hal_device_property_set_string (stordev, "storage.bus", "sata");
+				}
+				close (fd);
+			}
 		}
-		drive_id_close(did);
+#endif
 
 		snprintf (attr_path, SYSFS_PATH_MAX,
 			  "%s/device/type", sysfs_path);
@@ -1504,7 +1564,6 @@ block_class_pre_process (ClassDeviceHandler *self,
 					"disk");
 				break;
 			case 1:	/* Tape */
-				has_removable_media = TRUE;
 				hal_device_property_set_string (
 					stordev,
 					"storage.drive_type", 
@@ -1526,35 +1585,26 @@ block_class_pre_process (ClassDeviceHandler *self,
 			}
 		}
 
-		/* Check for (USB) floppy - Yuck, this is pretty ugly in
-		 * quite a few ways!! Reading /proc-files, looking at
-		 * some hardcoded string from the kernel etc. etc.
+		/* Check for USB floppy drive by looking at USB Mass Storage interface class
+		 * instead of Protocol: Uniform Floppy Interface (UFI) in /proc as we did before.
 		 *
-		 * @todo : Get Protocol into sysfs or something more sane
-		 * giving us the type of the SCSI device
+		 * (should fix RH bug 133834)
 		 */
-		scsi_host = hal_device_property_get_int (
-			parent, "scsi.host");
-		scsi_protocol = read_single_line_grep (
-			"     Protocol: ", 
-			"/proc/scsi/usb-storage/%d", 
-			scsi_host);
-		if (scsi_protocol != NULL &&
-		    strcmp (scsi_protocol,
-			    "Uniform Floppy Interface (UFI)") == 0) {
+		if (physdev != NULL) {
+			if (hal_device_property_get_int (physdev, "usb.interface.class") == 8 &&
+			    hal_device_property_get_int (physdev, "usb.interface.subclass") == 4 ) {
 
-			/* Indeed a (USB) floppy drive */
+				hal_device_property_set_string (d, "storage.drive_type", "floppy");
 
-			hal_device_property_set_string (d, 
-							"storage.drive_type", 
-							"floppy");
+				/* My experiments with my USB LaCie Floppy disk
+				 * drive is that polling indeed work (Yay!), so
+				 * we don't set storage.media_check_enabled to 
+				 * FALSE - for devices where this doesn't work,
+				 * we can override it with .fdi files
+				 */
+				has_removable_media = TRUE;
 
-			/* My experiments with my USB LaCie Floppy disk
-			 * drive is that polling indeed work (Yay!), so
-			 * we don't set storage.media_check_enabled to 
-			 * FALSE - for devices where this doesn't work,
-			 * we can override it with .fdi files
-			 */
+			}
 		}
 
 	} else {
@@ -1567,42 +1617,27 @@ block_class_pre_process (ClassDeviceHandler *self,
 		
 	}
 
-	snprintf (attr_path, SYSFS_PATH_MAX, "%s/removable", sysfs_path);
-	attr = sysfs_open_attribute (attr_path);
-	if (sysfs_read_attribute (attr) >= 0) {
-		if (attr->value [0] == '0')
-			has_removable_media = FALSE;
-		else
-			has_removable_media = TRUE;
-
-		sysfs_close_attribute (attr);
-	}
-
-
-	
 	hal_device_property_set_bool (stordev, "storage.removable", has_removable_media);
 
 	if (hal_device_has_property (stordev, "storage.drive_type") &&
-	    strcmp (hal_device_property_get_string (stordev, "storage.drive_type"), 
-		    "cdrom") == 0) {
+	    strcmp (hal_device_property_get_string (stordev, "storage.drive_type"), "cdrom") == 0) {
 		cdrom_get_properties (stordev, device_file);
 		hal_device_add_capability (stordev, "storage.cdrom");
 		hal_device_property_set_bool (stordev, "storage.no_partitions_hint", TRUE);
+		requires_eject = TRUE;
 	}
 
 
 	if (hal_device_has_property (stordev, "storage.drive_type") &&
-	    strcmp (hal_device_property_get_string (stordev, "storage.drive_type"), 
-		    "floppy") == 0) {
+	    strcmp (hal_device_property_get_string (stordev, "storage.drive_type"), "floppy") == 0) {
 		hal_device_property_set_bool (stordev, "storage.no_partitions_hint", TRUE);
 	}
 
 	hal_device_property_set_string (stordev, "info.category", "storage");
 	hal_device_add_capability (stordev, "storage");
 
-	hal_device_property_set_bool (stordev, "storage.hotpluggable",
-				      is_hotpluggable);
-
+	hal_device_property_set_bool (stordev, "storage.hotpluggable", is_hotpluggable);
+	hal_device_property_set_bool (stordev, "storage.requires_eject", requires_eject);
 	hal_device_property_set_bool (stordev, "block.no_partitions", 
 				      hal_device_property_get_bool (
 					      stordev, "storage.no_partitions_hint"));
@@ -1907,8 +1942,10 @@ etc_mtab_process_line (char *s)
 	mount_points[num_mount_points].minor = minor;
 	strncpy (mount_points[num_mount_points].device, device,
 		 MOUNT_POINT_STRING_SIZE);
-	strncpy (mount_points[num_mount_points].mount_point, mount_point,
-		 MOUNT_POINT_STRING_SIZE);
+	p = g_strcompress(mount_point);
+	strncpy (mount_points[num_mount_points].mount_point, p, 
+		MOUNT_POINT_STRING_SIZE);
+	g_free(p);
 	strncpy (mount_points[num_mount_points].fs_type, fs_type,
 		 MOUNT_POINT_STRING_SIZE);
 
@@ -1987,6 +2024,7 @@ static dbus_bool_t have_setup_watcher = FALSE;
 static gboolean
 mtab_handle_storage (HalDevice *d)
 {
+	const char *device_file;
 	HalDevice *child;
 	int major, minor;
 	dbus_bool_t found_mount_point;
@@ -2004,11 +2042,15 @@ mtab_handle_storage (HalDevice *d)
 
 	major = hal_device_property_get_int (d, "block.major");
 	minor = hal_device_property_get_int (d, "block.minor");
+	device_file = hal_device_property_get_string (d, "block.device");
+
 
 	/* See if we already got children */
 	child = get_child_device_gdl (d);
-	if (child == NULL)
-		get_child_device_tdl (d);
+
+	if (child == NULL) {
+		child = get_child_device_tdl (d);
+	}
 
 	/* Search all mount points */
 	found_mount_point = FALSE;
@@ -2016,8 +2058,13 @@ mtab_handle_storage (HalDevice *d)
 
 		mp = &mount_points[i];
 
-		if (mp->major != major || mp->minor != minor)
-			continue;
+		if (strcmp (device_file, mp->device) == 0) {
+			/* looks good */
+		} else {
+			/* device file don't match; try matching up major/minor numbers */
+			if (mp->major != major || mp->minor != minor)
+				continue;
+		}
 
 		if (child != NULL ) {
 			found_mount_point = TRUE;
@@ -2040,6 +2087,7 @@ mtab_handle_storage (HalDevice *d)
 static gboolean
 mtab_handle_volume (HalDevice *d)
 {
+	const char *device_file;
 	int major, minor;
 	dbus_bool_t found_mount_point;
 	struct mount_point_s *mp;
@@ -2049,6 +2097,7 @@ mtab_handle_volume (HalDevice *d)
 
 	major = hal_device_property_get_int (d, "block.major");
 	minor = hal_device_property_get_int (d, "block.minor");
+	device_file = hal_device_property_get_string (d, "block.device");
 
 	/* these are handled in mtab_handle_storage */
 	storudi = hal_device_property_get_string (d, "block.storage_device");
@@ -2062,36 +2111,27 @@ mtab_handle_volume (HalDevice *d)
 	found_mount_point = FALSE;
 	for (i = 0; i < num_mount_points; i++) {
 		mp = &mount_points[i];
-			
-		if (mp->major == major && mp->minor == minor) {
+
+	
+		/* match on both device file and major/minor numbers */
+		if ((strcmp (device_file, mp->device) == 0) || (mp->major == major && mp->minor == minor)) {
 			const char *existing_block_device;
 			dbus_bool_t was_mounted;
 
 			device_property_atomic_update_begin ();
 
-			existing_block_device =
-				hal_device_property_get_string (d,
-								"block.device");
+			existing_block_device = hal_device_property_get_string (d, "block.device");
 
-			was_mounted =
-				hal_device_property_get_bool (d,
-							      "volume.is_mounted");
+			was_mounted = hal_device_property_get_bool (d, "volume.is_mounted");
 
 			/* Yay! Found a mount point; set properties accordingly */
-			hal_device_property_set_string (d,
-							"volume.mount_point",
-							mp->mount_point);
-			hal_device_property_set_string (d, "volume.fstype",
-							mp->fs_type);
-			hal_device_property_set_bool (d,
-						      "volume.is_mounted",
-						      TRUE);
+			hal_device_property_set_string (d, "volume.mount_point", mp->mount_point);
+			hal_device_property_set_string (d, "volume.fstype", mp->fs_type);
+			hal_device_property_set_bool (d, "volume.is_mounted", TRUE);
 
 			/* only overwrite block.device if it's not set */
-			if (existing_block_device == NULL ||
-			    (existing_block_device != NULL &&
-			     strcmp (existing_block_device,
-				     "") == 0)) {
+			if (existing_block_device == NULL || (existing_block_device != NULL &&
+			     strcmp (existing_block_device, "") == 0)) {
 				hal_device_property_set_string (d,
 								"block.device",
 								mp->
@@ -2206,7 +2246,7 @@ mtab_foreach_device (HalDeviceStore *store, HalDevice *d,
  *
  *  @param  force               Force reading of mtab
  */
-static void
+void
 etc_mtab_process_all_block_devices (dbus_bool_t force)
 {
 	/* Start or continue watching /etc */
@@ -2251,16 +2291,39 @@ sigio_handler (int sig)
 }
 
 
-static void
+static dbus_bool_t
 block_class_removed (ClassDeviceHandler* self, 
 		     const char *sysfs_path, 
 		     HalDevice *d)
 {
 	HalDevice *child;
 
+	HAL_INFO (("entering, sysfs_path=%s", sysfs_path));
+
 	if (hal_device_has_property (d, "block.is_volume")) {
 		if (hal_device_property_get_bool (d, "block.is_volume")) {
+			const char *storudi;
+			HalDevice *stordev;
+			const char *storbus;
+
+			storudi = hal_device_property_get_string (d, "block.storage_device");
+			if (storudi != NULL) {
+				stordev = hal_device_store_find (hald_get_gdl (), storudi);
+				if (stordev != NULL) {
+					storbus = hal_device_property_get_string (stordev, "storage.bus");
+					if (storbus != NULL) {
+						if (strcmp (storbus, "ide") == 0) {
+							HAL_INFO (("Explicitly ignoring IDE block device remove for %s",
+								   sysfs_path));
+							return FALSE;
+						}
+					}
+				}
+			}
+
+
 			force_unmount (d);
+
 		} else {
 			force_unmount_of_all_childs (d);
 
@@ -2274,13 +2337,43 @@ block_class_removed (ClassDeviceHandler* self,
 
 				if (child != NULL) {
 					g_signal_connect (child, "callouts_finished",
-							  G_CALLBACK (volume_remove_from_gdl), NULL);
+							  G_CALLBACK (volume_remove_from_gdl_sleep), NULL);
 					hal_callout_device (child, FALSE);
 				}
+			} else {
+				GSList *children;
+
+				/* ensure all childs are gone (only applicable for block devices from IDE since 
+				 * they won't be deleted on hotplug remove we don't want that, cf. above)
+				 */
+				children = hal_device_store_match_multiple_key_value_string (
+					hald_get_gdl (),
+					"info.parent",
+					hal_device_get_udi (d));
+				
+				if (children != NULL) {
+					GSList *iter;
+
+					for (iter = children; iter != NULL; iter = iter->next) {
+						HalDevice *child = HAL_DEVICE (iter->data);
+
+						HAL_INFO (("Dumping child %s", child->udi));
+						if (child != NULL) {
+							g_signal_connect (child, "callouts_finished",
+									  G_CALLBACK (volume_remove_from_gdl_sleep), NULL);
+							hal_callout_device (child, FALSE);
+						}
+					}
+
+					g_slist_free (children);
+				}
+
 			}
 
 		}
 	}
+
+	return TRUE;
 }
 
 
