@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: class_device.c,v 1.28 2004/08/28 13:35:47 david Exp $
+ * CVSID: $Id: class_device.c,v 1.29.2.1 2004/12/01 03:39:01 david Exp $
  *
  * Generic methods for class devices
  *
@@ -202,15 +202,11 @@ class_device_visit (ClassDeviceHandler *self,
 
 	} else {
 		if (sysdevice != NULL) {
-			what_to_find = get_parent_sysfs_path (sysdevice->path);
+			hal_target = find_closest_ancestor (sysdevice->path);
 		} else {
-			what_to_find = "(none)"; /* DZE: Yikes, hack to get added as a child to 'computer'? */
+			hal_target = find_computer ();
 		}
 
-		/* find the sysdevice we should be a child of */
-		hal_target = hal_device_store_match_key_value_string (hald_get_gdl (), 
-								      "linux.sysfs_path_device", 
-								      what_to_find);
 		if (hal_target == NULL)
 			goto error;
 
@@ -233,11 +229,12 @@ error:
  *  @param  d                  The HalDevice object of the instance of
  *                             this device class
  */
-void
+dbus_bool_t
 class_device_removed (ClassDeviceHandler* self, const char *sysfs_path, 
 		      HalDevice *d)
 {
 	HAL_INFO (("sysfs_path = '%s'", sysfs_path));
+	return TRUE;
 }
 
 /** Called when a device file (e.g. a file in /dev) have been created by udev 
