@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: net_class_device.c,v 1.18 2004/09/16 22:04:15 david Exp $
+ * CVSID: $Id: net_class_device.c,v 1.21.2.4 2005/01/17 20:00:30 david Exp $
  *
  * Network device class
  *
@@ -58,6 +58,10 @@
 #include "class_device.h"
 #include "common.h"
 
+#if PCMCIA_SUPPORT_ENABLE
+#include "pcmcia_utils.h"
+#endif
+
 /**
  * @defgroup HalDaemonLinuxNet Network class
  * @ingroup HalDaemonLinux
@@ -143,11 +147,13 @@ mii_get_rate (HalDevice *d)
 {
 	const char *ifname;
 	int sockfd;
-	struct ifreq ifr;
+	struct ifreq *ifr;
 	gboolean new_ioctl_nums;
 	int res;
 	guint16 link_word;
 
+	HAL_INFO (("Entering"));
+
 	ifname = hal_device_property_get_string (d, "net.interface");
 
 	sockfd = socket (AF_INET, SOCK_DGRAM, 0);
@@ -157,16 +163,20 @@ mii_get_rate (HalDevice *d)
 		return;
 	}
 
-	snprintf (ifr.ifr_name, IFNAMSIZ, ifname);
+	/* reserve some extra space as the sk98lin driver seems to segfault otherwise */
+	ifr = g_new0 (struct ifreq, 10);
 
-	if (ioctl (sockfd, 0x8947, &ifr) >= 0)
+	snprintf (ifr->ifr_name, IFNAMSIZ, ifname);
+
+	if (ioctl (sockfd, 0x8947, ifr) >= 0)
 		new_ioctl_nums = TRUE;
-	else if (ioctl (sockfd, SIOCDEVPRIVATE, &ifr) >= 0)
+	else if (ioctl (sockfd, SIOCDEVPRIVATE, ifr) >= 0)
 		new_ioctl_nums = FALSE;
 	else {
 		HAL_ERROR (("SIOCGMIIPHY on %s failed: %s",
-			    ifr.ifr_name, strerror (errno)));
+			    ifr->ifr_name, strerror (errno)));
 		close (sockfd);
+		g_free (ifr);
 		return;
 	}
 
@@ -183,7 +193,7 @@ mii_get_rate (HalDevice *d)
 	 * 0x0020  10baseT supported
 	 * 0x001F  Protocol selection bits, always 0x0001 for Ethernet.
 	 */
-	res = mdio_read (sockfd, &ifr, 5, new_ioctl_nums, &link_word);
+	res = mdio_read (sockfd, ifr, 5, new_ioctl_nums, &link_word);
 
 	if (res < 0) {
 		HAL_WARNING (("Error reading rate info"));
@@ -196,6 +206,7 @@ mii_get_rate (HalDevice *d)
 	}
 
 	close (sockfd);
+	g_free (ifr);
 }
 
 static void
@@ -203,11 +214,13 @@ mii_get_link (HalDevice *d)
 {
 	const char *ifname;
 	int sockfd;
-	struct ifreq ifr;
+	struct ifreq *ifr;
 	gboolean new_ioctl_nums;
 	int res;
 	guint16 status_word;
 
+	HAL_INFO (("Entering"));
+
 	ifname = hal_device_property_get_string (d, "net.interface");
 
 	sockfd = socket (AF_INET, SOCK_DGRAM, 0);
@@ -217,16 +230,20 @@ mii_get_link (HalDevice *d)
 		return;
 	}
 
-	snprintf (ifr.ifr_name, IFNAMSIZ, ifname);
+	/* reserve some extra space as the sk98lin driver seems to segfault otherwise */
+	ifr = g_new0 (struct ifreq, 10);
 
-	if (ioctl (sockfd, 0x8947, &ifr) >= 0)
+	snprintf (ifr->ifr_name, IFNAMSIZ, ifname);
+
+	if (ioctl (sockfd, 0x8947, ifr) >= 0)
 		new_ioctl_nums = TRUE;
-	else if (ioctl (sockfd, SIOCDEVPRIVATE, &ifr) >= 0)
+	else if (ioctl (sockfd, SIOCDEVPRIVATE, ifr) >= 0)
 		new_ioctl_nums = FALSE;
 	else {
 		HAL_ERROR (("SIOCGMIIPHY on %s failed: %s",
-			    ifr.ifr_name, strerror (errno)));
+			    ifr->ifr_name, strerror (errno)));
 		close (sockfd);
+		g_free (ifr);
 		return;
 	}
 
@@ -246,8 +263,8 @@ mii_get_link (HalDevice *d)
 	 */
 
 	/* We have to read it twice to clear any "sticky" bits */
-	res = mdio_read (sockfd, &ifr, 1, new_ioctl_nums, &status_word);
-	res = mdio_read (sockfd, &ifr, 1, new_ioctl_nums, &status_word);
+	res = mdio_read (sockfd, ifr, 1, new_ioctl_nums, &status_word);
+	res = mdio_read (sockfd, ifr, 1, new_ioctl_nums, &status_word);
 
 	if (res < 0)
 		HAL_WARNING (("Error reading link info"));
@@ -256,10 +273,12 @@ mii_get_link (HalDevice *d)
 	else
 		hal_device_property_set_bool (d, "net.80203.link", FALSE);
 
+	close (sockfd);
+	g_free (ifr);
+
 	/* Also get the link rate */
 	mii_get_rate (d);
 
-	close (sockfd);
 }
 
 static void
@@ -325,7 +344,12 @@ link_detection_handle_message (struct nlmsghdr *hdr)
 			if (hal_device_has_capability (d, "net.80203")) {
 				if (!hal_device_property_get_bool (d, "net.80203.link")) {
 					hal_device_property_set_bool (d, "net.80203.link", TRUE);
+#ifdef SYSFS_CARRIER_ENABLE
+					HAL_INFO (("Assuming link speed is 100Mbps"));
+					hal_device_property_set_uint64 (d, "net.80203.rate", 100 * 1000 * 1000);
+#else /* SYSFS_CARRIER_ENABLE */
 					mii_get_rate (d);
+#endif /* SYSFS_CARRIER_ENABLE */
 				}
 			}
 		} else {
@@ -387,6 +411,8 @@ link_detection_data_ready (GIOChannel *channel, GIOCondition cond,
 	int fd;
 	int bytes_read;
 	guint total_read = 0;
+	struct sockaddr_nl nladdr;
+	socklen_t nladdrlen = sizeof(nladdr);
 	char buf[1024];
 
 	if (cond & ~(G_IO_IN | G_IO_PRI)) {
@@ -398,11 +424,19 @@ link_detection_data_ready (GIOChannel *channel, GIOCondition cond,
 
 	do {
 		errno = 0;
-		bytes_read = recv (fd,
+		bytes_read = recvfrom (fd,
 				   buf + total_read,
 				   sizeof (buf) - total_read,
-				   MSG_DONTWAIT);
-
+				   MSG_DONTWAIT,
+				   (struct sockaddr*)&nladdr, &nladdrlen);
+		if (nladdrlen != sizeof(nladdr)) {
+			HAL_ERROR(("Bad address size reading netlink socket"));
+			return FALSE;
+		}
+		if (nladdr.nl_pid) {
+			HAL_ERROR(("Spoofed packet received on netlink socket"));
+			return FALSE;
+		}
 		if (bytes_read > 0)
 			total_read += bytes_read;
 	} while (bytes_read > 0 || errno == EINTR);
@@ -499,30 +533,31 @@ net_class_pre_process (ClassDeviceHandler *self,
 	int media_type = 0;
 	const char *media;
 	char wireless_path[SYSFS_PATH_MAX];
-	struct sysfs_directory *wireless_dir;
-	dbus_bool_t is_80211 = FALSE;
+	char driver_path[SYSFS_PATH_MAX];
+	dbus_bool_t is_80211;
 	int ifindex;
 	int flags;
+	struct stat statbuf;
+#if PCMCIA_SUPPORT_ENABLE
+ 	pcmcia_stab_entry *entry;
+#endif
 
 	hal_device_property_set_string (d, "net.linux.sysfs_path", sysfs_path);
 	hal_device_property_set_string (d, "net.interface",
 					class_device->name);
 
-	/* Check to see if this interface supports wireless extensions */
 	is_80211 = FALSE;
-	snprintf (wireless_path, SYSFS_PATH_MAX, "%s/wireless", sysfs_path);
-	wireless_dir = sysfs_open_directory (wireless_path);
 
-	if (sysfs_read_directory (wireless_dir) >= 0) {
-		if (wireless_dir->attributes != NULL) {
-			hal_device_add_capability (d, "net.80211");
-			is_80211 = TRUE;
+	/* Check driver link (may be unavailable for PCMCIA devices) */
+	snprintf (driver_path, SYSFS_PATH_MAX, "%s/driver", sysfs_path);
+	if (stat (driver_path, &statbuf) == 0) {
+		char buf[256];
+		memset (buf, '\0', sizeof (buf));
+		if (readlink (driver_path, buf, sizeof (buf) - 1) > 0) {
+			hal_device_property_set_string (d, "net.linux.driver", get_last_element (buf));
 		}
-
-		sysfs_close_directory (wireless_dir);
 	}
 
-
 	attr = sysfs_get_classdev_attr (class_device, "address");
 	if (attr != NULL) {
 		address = g_strstrip (g_strdup (attr->value));
@@ -534,14 +569,77 @@ net_class_pre_process (ClassDeviceHandler *self,
 		media_type = parse_dec (attr->value);
 	}
 
+	if (media_type == ARPHRD_ETHER) {
+		FILE *f;
+		dbus_bool_t is_wireless;
+
+
+		is_wireless = FALSE;
+
+		f = fopen ("/proc/net/wireless", "ro");
+		if (f != NULL) {
+			unsigned int i;
+			unsigned int ifname_len;
+			char buf[128];
+
+			ifname_len = strlen (class_device->name);
+
+			do {
+				if (fgets (buf, sizeof (buf), f) == NULL)
+					break;
+
+				for (i=0; i < sizeof (buf); i++) {
+					if (isspace (buf[i]))
+						continue;
+					else
+						break;
+				}
+
+				if (strncmp (class_device->name, buf + i, ifname_len) == 0) {
+					is_wireless = TRUE;
+					break;
+				}
+
+			} while (TRUE);
+			fclose (f);
+		}
+
+		if (is_wireless) {
+		/* Check to see if this interface supports wireless extensions */
+		/*
+		snprintf (wireless_path, SYSFS_PATH_MAX, "%s/wireless", sysfs_path);
+		if (stat (wireless_path, &statbuf) == 0) {
+		*/
+			hal_device_property_set_string (d, "info.category", "net.80211");
+			hal_device_add_capability (d, "net.80211");
+			is_80211 = TRUE;
+		} else {
+			hal_device_property_set_string (d, "info.category", "net.80203");
+			hal_device_add_capability (d, "net.80203");
+		}
+	}
+
 	attr = sysfs_get_classdev_attr (class_device, "flags");
 	if (attr != NULL) {
 		flags = parse_hex (attr->value);
 		hal_device_property_set_bool (d, "net.interface_up", flags & IFF_UP);
-		if (!is_80211) {
+		if (!is_80211 && media_type == ARPHRD_ETHER) {
 			/* TODO: for some reason IFF_RUNNING isn't exported in flags */
 			/*hal_device_property_set_bool (d, "net.80203.link", flags & IFF_RUNNING);*/
+#ifdef SYSFS_CARRIER_ENABLE
+			attr = sysfs_get_classdev_attr (class_device, "carrier");
+			if (attr != NULL) {
+				int have_link;
+
+				have_link = parse_dec (attr->value);
+				HAL_INFO (("According to sysfs link status is %d", have_link));
+				hal_device_property_set_bool (d, "net.80203.link", have_link != 0);
+				HAL_INFO (("Assuming link speed is 100Mbps"));
+				hal_device_property_set_uint64 (d, "net.80203.rate", 100 * 1000 * 1000);
+			}
+#else /* SYSFS_CARRIER_ENABLE */
 			mii_get_link (d);
+#endif /* SYSFS_CARRIER_ENABLE */
 		}
 	}
 
@@ -577,7 +675,12 @@ net_class_pre_process (ClassDeviceHandler *self,
 
 	if (hal_device_has_property (d, "net.80203.link") &&
 	    hal_device_property_get_bool (d, "net.80203.link")) {
+#ifdef SYSFS_CARRIER_ENABLE
+		HAL_INFO (("Assuming link speed is 100Mbps"));
+		hal_device_property_set_uint64 (d, "net.80203.rate", 100 * 1000 * 1000);
+#else /* SYSFS_CARRIER_ENABLE */
 		mii_get_rate (d);
+#endif /* SYSFS_CARRIER_ENABLE */
 	}
 
 	hal_device_property_set_int (d, "net.arp_proto_hw_id", media_type);
@@ -586,13 +689,69 @@ net_class_pre_process (ClassDeviceHandler *self,
 	hal_device_property_set_string (d, "net.media", media);
 
 	hal_device_add_capability (d, "net");
-	if (is_80211) {
-		hal_device_property_set_string (d, "info.category", "net.80211");
-		hal_device_add_capability (d, "net.80211");
-	} else {
-		hal_device_property_set_string (d, "info.category", "net.80203");
-		hal_device_add_capability (d, "net.80203");
+
+#if PCMCIA_SUPPORT_ENABLE
+	/* Add PCMCIA specific entries for PCMCIA cards */
+	if ((entry = pcmcia_get_stab_entry_for_device (class_device->name))) {
+		pcmcia_card_info *info = pcmcia_card_info_get (entry->socket);
+		if (info && (info->socket >= 0)) {
+			const char *type;
+			HalDevice *parent;
+
+			hal_device_property_set_string (d, "info.bus", "pcmcia");
+			if (entry->driver)
+				hal_device_property_set_string (d, "net.linux.driver", entry->driver);
+
+			if (info->productid_1 && strlen (info->productid_1))
+				hal_device_property_set_string (d, "pcmcia.productid_1", info->productid_1);
+			if (info->productid_2 && strlen (info->productid_2))
+				hal_device_property_set_string (d, "pcmcia.productid_2", info->productid_2);
+			if (info->productid_3 && strlen (info->productid_3))
+				hal_device_property_set_string (d, "pcmcia.productid_3", info->productid_3);
+			if (info->productid_4 && strlen (info->productid_4))
+				hal_device_property_set_string (d, "pcmcia.productid_4", info->productid_4);
+
+			if ((type = pcmcia_card_type_string_from_type (info->type)))
+				hal_device_property_set_string (d, "pcmcia.function", type);
+
+			hal_device_property_set_int (d, "pcmcia.manfid_1", info->manfid_1);
+			hal_device_property_set_int (d, "pcmcia.manfid_2", info->manfid_2);
+			hal_device_property_set_int (d, "pcmcia.socket_number", info->socket);
+
+			/* Provide best-guess of vendor, goes in Vendor property; 
+			 * .fdi files can override this */
+			if (info->productid_1 != NULL) {
+				hal_device_property_set_string (d, "info.vendor", info->productid_1);
+			} else {
+				char namebuf[50];
+				snprintf (namebuf, sizeof(namebuf), "Unknown (0x%04x)", info->manfid_1);
+				hal_device_property_set_string (d, "info.vendor", namebuf);
+			}
+
+			/* Provide best-guess of name, goes in Product property; 
+			 * .fdi files can override this */
+			if (info->productid_2 != NULL) {
+				hal_device_property_set_string (d, "info.product", info->productid_2);
+			} else {
+				char namebuf[50];
+				snprintf (namebuf, sizeof(namebuf), "Unknown (0x%04x)", info->manfid_2);
+				hal_device_property_set_string (d, "info.product", namebuf);
+			}
+
+			/* Reparent PCMCIA devices to be under their socket */
+			parent = hal_device_store_match_key_value_int (hald_get_gdl (), 
+									      "pcmcia_socket.number", 
+									      info->socket);
+			if (parent)
+				hal_device_property_set_string (d, "info.parent",
+						hal_device_property_get_string (parent, "info.udi"));
+
+		}
+
+		pcmcia_card_info_free (info);
+		pcmcia_stab_entry_free (entry);
 	}
+#endif
 }
 
 static dbus_bool_t
@@ -601,6 +760,9 @@ net_class_accept (ClassDeviceHandler *self, const char *path,
 {
 	gboolean accept;
 	struct sysfs_attribute *attr = NULL;
+#if PCMCIA_SUPPORT_ENABLE
+	pcmcia_stab_entry *entry = NULL;
+#endif
 
 	accept = TRUE;
 
@@ -627,12 +789,32 @@ net_class_accept (ClassDeviceHandler *self, const char *path,
 	}
 
 	/* 1 is the type for ethernet (incl. wireless) devices */
-	if (attr->value[0] == '1') {
-		accept = TRUE;
+	if (attr->value[0] != '1') {
+		accept = FALSE;
+		goto out;
+	}
+
+#if PCMCIA_SUPPORT_ENABLE
+	/* Allow 'net' devices without a sysdevice only if they
+	 * are backed by real hardware (like PCMCIA cards).
+	 */
+	if ((entry = pcmcia_get_stab_entry_for_device (class_device->name))) {
+		pcmcia_card_info *info = pcmcia_card_info_get (entry->socket);
+		if (info && (info->socket >= 0)) {
+			/* Ok, we're a PCMCIA card */
+			accept = TRUE;
+		} else {
+			accept = FALSE;
+		}
+		
+		pcmcia_card_info_free (info);
+		pcmcia_stab_entry_free (entry);
 		goto out;
 	} else {
 		accept = FALSE;
+		goto out;
 	}
+#endif
 
 out:
 	return accept;
@@ -655,7 +837,7 @@ net_class_compute_udi (HalDevice *d, int append_num)
 	if (append_num == -1)
 		format = "/org/freedesktop/Hal/devices/net-%s";
 	else
-		format = "/org/freedesktop/Hal/devices_net-%s-%d";
+		format = "/org/freedesktop/Hal/devices/net-%s-%d";
 
 	snprintf (buf, 256, format,
 		  hal_device_property_get_string (d, "net.address"),
