@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: osspec.c,v 1.43 2004/09/16 22:04:15 david Exp $
+ * CVSID: $Id: osspec.c,v 1.52.2.5 2005/01/21 15:46:55 david Exp $
  *
  * Detection and monitoring of devices on Linux 2.6 + udev
  *
@@ -37,10 +37,13 @@
 #include <unistd.h>
 #include <stdarg.h>
 #include <errno.h>
+#include <syslog.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/utsname.h>
+#include <sys/stat.h>
+#include <time.h>
 
 #include <glib.h>
 #include <dbus/dbus.h>
@@ -50,6 +53,8 @@
 #include "../logger.h"
 #include "../hald.h"
 #include "../callout.h"
+#include "../device_info.h"
+#include "../hald_conf.h"
 
 #include "common.h"
 #include "hald_helper.h"
@@ -58,6 +63,12 @@
 
 #include "libsysfs/libsysfs.h"
 
+/** How many ms to sleep on first hotplug event (to queue up other hotplug events) */
+#define FIRST_HOTPLUG_SLEEP 3500
+
+/** How many seconds before we discard a missing hotplug event and move on to the next one */
+#define HOTPLUG_TIMEOUT 25
+
 extern ClassDeviceHandler input_class_handler;
 extern ClassDeviceHandler net_class_handler;
 extern ClassDeviceHandler printer_class_handler;
@@ -69,6 +80,7 @@ extern ClassDeviceHandler pcmcia_socket_class_handler;
 extern ClassDeviceHandler ieee1394_class_handler;
 extern ClassDeviceHandler ieee1394_node_class_handler;
 extern ClassDeviceHandler ieee1394_host_class_handler;
+extern ClassDeviceHandler serial_class_handler;
 extern ClassDeviceHandler multimedia_class_handler;
 
 extern BusDeviceHandler pci_bus_handler;
@@ -78,7 +90,9 @@ extern BusDeviceHandler ide_host_bus_handler;
 extern BusDeviceHandler ide_bus_handler;
 extern BusDeviceHandler scsi_bus_handler;
 extern BusDeviceHandler macio_bus_handler;
+extern BusDeviceHandler pcmcia_bus_handler;
 extern BusDeviceHandler platform_bus_handler;
+extern BusDeviceHandler usb_serial_bus_handler;
 
 /*
  * NOTE!  Order can be significant here, especially at startup time
@@ -98,6 +112,7 @@ static ClassDeviceHandler* class_device_handlers[] = {
 	&ieee1394_host_class_handler,
 	&ieee1394_node_class_handler,
 	&ieee1394_class_handler,
+	&serial_class_handler,
 	&multimedia_class_handler,
 	NULL
 };
@@ -107,10 +122,12 @@ static BusDeviceHandler* bus_device_handlers[] = {
 	&usb_bus_handler,
 	&usbif_bus_handler,
 	&ide_host_bus_handler,
-        &ide_bus_handler,
+	&ide_bus_handler,
 	&macio_bus_handler,
+	&pcmcia_bus_handler,
 	&platform_bus_handler,
 	&scsi_bus_handler,
+	&usb_serial_bus_handler,
 	NULL
 };
 
@@ -125,6 +142,8 @@ static gboolean hald_helper_data (GIOChannel *source, GIOCondition condition, gp
 
 static HalDevice *add_device (const char *sysfs_path, const char *subsystem, struct hald_helper_msg *msg);
 
+static void hotplug_timeout_handler (void);
+
 /** Mount path for sysfs */
 char sysfs_mount_path[SYSFS_PATH_MAX];
 
@@ -148,6 +167,8 @@ osspec_timer_handler (gpointer data)
 		ch->tick (ch);
 	}
 
+	hotplug_timeout_handler ();
+
 	return TRUE;
 }
 
@@ -440,10 +461,23 @@ compute_coldplug_list (void)
 			gchar *normalized_target;
 
 			g_snprintf (path, SYSFS_PATH_MAX, "%s/class/%s/%s/device", sysfs_mount_path, f, f1);
-			if ((target = g_file_read_link (path, NULL)) != NULL) {
+			/* Accept net devices without device links too, they may be coldplugged PCMCIA devices */
+			if (((target = g_file_read_link (path, NULL)) != NULL)
+#if PCMCIA_SUPPORT_ENABLE
+			    || !strcmp (f, "net")
+#endif
+				)
+			{
 				g_snprintf (path1, SYSFS_PATH_MAX, "%s/class/%s/%s", sysfs_mount_path, f, f1);
-				normalized_target = get_normalized_path (path1, target);
-				g_free (target);
+				if (target) {
+					normalized_target = get_normalized_path (path1, target);
+					g_free (target);
+				}
+#if PCMCIA_SUPPORT_ENABLE
+				else {
+					normalized_target = g_strdup (path1);
+				}
+#endif
 
 				/*printf ("%s -> (%s, %s)\n", normalized_target, path1, f);*/
 				g_hash_table_insert (sysfs_to_class_in_devices_map, 
@@ -599,7 +633,7 @@ compute_coldplug_visit_device (const gchar *path, GSList **coldplug_list)
 	GError *err;
 	GDir *dir;
 	const gchar *f;
-		
+
 	bus = g_hash_table_lookup (sysfs_to_bus_map, path);
 	if (bus != NULL) {
 		for (i = 0; bus_device_handlers[i] != NULL; i++) {	
@@ -650,22 +684,29 @@ compute_coldplug_visit_device (const gchar *path, GSList **coldplug_list)
 						 h_string_pair_new (g_strdup (path), g_strdup (bus)));
 		goto found;
 	}
-	return;
 
 found:
 
 	/* visit children */
+	err = NULL;
 	if ((dir = g_dir_open (path, 0, &err)) == NULL) {
-		HAL_ERROR (("Unable to open %: %s", path, err->message));
+		/*HAL_ERROR (("Unable to open directory: %s", path, err->message));*/
 		g_error_free (err);
 		goto error;
 	}
 	while ((f = g_dir_read_name (dir)) != NULL) {
 		gchar path_child[SYSFS_PATH_MAX];
-		
+		struct stat statbuf;
+	
 		g_snprintf (path_child, SYSFS_PATH_MAX, "%s/%s", path, f);
-		
-		compute_coldplug_visit_device (path_child, coldplug_list);
+
+		if (lstat (path_child, &statbuf) == 0) {
+
+			if (!S_ISLNK (statbuf.st_mode)) {
+				/* recursion fun */
+				compute_coldplug_visit_device (path_child, coldplug_list);
+			}
+		}
 	}
 	g_dir_close (dir);
 
@@ -795,6 +836,47 @@ add_computer_callouts_done (HalDevice *device, gpointer user_data)
 	process_coldplug_list ();
 }
 
+#ifdef HAVE_SELINUX
+#include <selinux/selinux.h>
+
+#if 0
+static int get_selinux_removable_context(security_context_t *newcon)
+{
+	FILE *fp;
+	char buf[255], *ptr;
+	size_t plen;
+	
+	HAL_INFO (("selinux_removable_context_path '%s'", selinux_removable_context_path()));
+	fp = fopen(selinux_removable_context_path(), "r");
+	if (!fp)
+		return -1;
+	
+	ptr = fgets_unlocked(buf, sizeof buf, fp);
+	fclose(fp);
+	
+	if (!ptr)
+		return -1;
+	plen = strlen(ptr);
+	if (buf[plen-1] == '\n') 
+		buf[plen-1] = 0;
+	
+	*newcon=strdup(buf);
+	/* If possible, check the context to catch
+	   errors early rather than waiting until the
+	   caller tries to use setexeccon on the context.
+	   But this may not always be possible, e.g. if
+	   selinuxfs isn't mounted. */
+	if (security_check_context(*newcon) && errno != ENOENT) {
+		free(*newcon);
+		*newcon = 0;
+		return -1;
+	}
+	
+	HAL_INFO (("removable context is %s", *newcon));
+	return 0;
+}
+#endif
+#endif /* HAVE_SELINUX */
 
 /* This function is documented in ../osspec.h */
 void
@@ -821,15 +903,42 @@ osspec_probe (void)
 	hal_device_set_udi (root, "/org/freedesktop/Hal/devices/computer");
 	
 	if (uname (&un) >= 0) {
-		hal_device_property_set_string (root, "kernel.name",
-						un.sysname);
-		hal_device_property_set_string (root, "kernel.version",
-						un.release);
-		hal_device_property_set_string (root, "kernel.machine",
-						un.machine);
+		hal_device_property_set_string (root, "kernel.name", un.sysname);
+		hal_device_property_set_string (root, "kernel.version", un.release);
+		hal_device_property_set_string (root, "kernel.machine", un.machine);
 	}
+
+#ifdef HAVE_SELINUX
+	if (is_selinux_enabled()) {
+/*
+		char buf[256];
+		security_context_t scontext;
+*/
+		hal_device_property_set_bool (root, "linux.is_selinux_enabled", TRUE);
+
+/*
+		if (get_selinux_removable_context(&scontext)==0) {
+			snprintf (buf, sizeof (buf), "storage.policy.default.mount_option.fscontext=%s", scontext);
+			freecon(scontext);
+			hal_device_property_set_bool (root, buf, TRUE);
+		} else {
+			HAL_ERROR (("Could not get selinux removable fscontext"));
+		}
+*/
+	}
+#endif /* HAVE_SELINUX */
+
 	hal_device_store_add (hald_get_tdl (), root);
 
+	/* Search for device information file and attempt merge */
+	if (di_search_and_merge (root)) {
+		HAL_INFO (("Found a .fdi file for %s", root->udi));
+	}
+	
+	/* add possible saved properties for this udi from disk*/
+	if (hald_get_conf ()->persistent_device_list)
+		hal_pstore_load_device (hald_get_pstore_sys (), root);
+
 	/* begin processing the coldplug_list when computer is added */
 	g_signal_connect (root,
 			  "callouts_finished",
@@ -894,6 +1003,11 @@ add_device (const char *given_sysfs_path, const char *subsystem, struct hald_hel
 	char sysfs_path[SYSFS_PATH_MAX];
 	HalDevice *hal_device = NULL;
 
+	if (given_sysfs_path == NULL) {
+		HAL_WARNING (("given_sysfs_path is NULL, cannot add this device" ));
+		return NULL;
+	}
+
 	strncpy (sysfs_path, given_sysfs_path, SYSFS_PATH_MAX);
 
 	len1 = snprintf (buf1, SYSFS_PATH_MAX, "%s/block", sysfs_mount_path);
@@ -925,8 +1039,10 @@ add_device (const char *given_sysfs_path, const char *subsystem, struct hald_hel
 						} else {
 							HAL_INFO (("%s worked!", sysfs_path));
 						}
+					} else {
+						HAL_INFO (("Recovery failed!"));
+						return NULL;
 					}
-
 				} else {
 					HAL_WARNING (("Coulnd't get sysfs class device object at path %s", sysfs_path));
 					return NULL;
@@ -1005,20 +1121,25 @@ rem_device (const char *sysfs_path, const char *subsystem, struct hald_helper_ms
 			HAL_WARNING (("Removal of class device at sysfs path %s is not yet implemented", sysfs_path));
 			goto out;
 		} else {
+			dbus_bool_t really_remove = TRUE;
 
 			for (i=0; class_device_handlers[i] != NULL; i++) {
 				ClassDeviceHandler *ch = class_device_handlers[i];
 				/** @todo TODO FIXME: just use ->accept() once we get rid of libsysfs */
 				if (strcmp (ch->hal_class_name, subsystem) == 0) {
-					ch->removed (ch, sysfs_path, hal_device);
+					really_remove = ch->removed (ch, sysfs_path, hal_device);
 				}
 
 			}
 
-			g_signal_connect (hal_device, "callouts_finished",
-					  G_CALLBACK (rem_device_callouts_finished), NULL);
-			HAL_INFO (("in remove_device for udi=%s", hal_device->udi));
-			hal_callout_device (hal_device, FALSE);
+			if (really_remove) {
+				g_signal_connect (hal_device, "callouts_finished",
+						  G_CALLBACK (rem_device_callouts_finished), NULL);
+				HAL_INFO (("in remove_device for udi=%s", hal_device->udi));
+				hal_callout_device (hal_device, FALSE);
+			} else {
+				hal_device = NULL;
+			}
 			goto out;
 		}
 	} else {
@@ -1307,9 +1428,15 @@ hald_helper_device_name (gchar *action, guint64 seqnum, gchar *subsystem,
 /** queue of hotplug events (struct hald_helper_msg pointers) */
 static GList *hotplug_queue = NULL;
 
+/** queue of hotplug events received when sleeping on the first hotplug event (struct hald_helper_msg pointers) */
+static GList *hotplug_queue_first = NULL;
+
 /** Last hotplug sequence number */
 static guint64 last_hotplug_seqnum = 0;
 
+/** Timestamp of last hotplug */
+static time_t last_hotplug_time_stamp = 0;
+
 /** Hotplug semaphore */
 static gint hotplug_counter = 0;
 
@@ -1323,6 +1450,17 @@ trynext:
 	if (hotplug_counter > 0)
 		return;
 
+	/* Empty the list of events received while sleeping on the first hotplug event (this list is sorted) */
+	if (hotplug_queue_first != NULL) {
+		msg = (struct hald_helper_msg *) hotplug_queue_first->data;
+		HAL_INFO (("Processing event around first hotplug with SEQNUM=%llu", msg->seqnum));
+		hald_helper_hotplug (msg->action, msg->seqnum, g_strdup (msg->subsystem), 
+				     g_strdup (msg->sysfs_path), msg);
+		g_free (msg);
+		hotplug_queue_first = g_list_delete_link (hotplug_queue_first, hotplug_queue_first);
+		goto trynext;
+	}
+
 	for (i = hotplug_queue; i != NULL; i = g_list_next (i)) {
 		msg = (struct hald_helper_msg *) i->data;
 
@@ -1340,6 +1478,7 @@ trynext:
 		if (msg->seqnum == last_hotplug_seqnum + 1) {
 			/* yup, found it */
 			last_hotplug_seqnum = msg->seqnum;
+			last_hotplug_time_stamp = msg->time_stamp;
 			hald_helper_hotplug (msg->action, msg->seqnum, g_strdup (msg->subsystem), 
 					     g_strdup (msg->sysfs_path), msg);
 			g_free (msg);
@@ -1349,6 +1488,49 @@ trynext:
 	}
 }
 
+/** Check the queue and do timeout handling on missing hotplug events */
+static void 
+hotplug_timeout_handler (void)
+{
+	time_t now;
+	GList *i;
+	struct hald_helper_msg *msg;
+
+	now = time (NULL);
+
+	/* See if there was a last hotplug event */
+	if (last_hotplug_time_stamp > 0) {
+		/* See if it's too long ago we processed */
+		if (now - last_hotplug_time_stamp > HOTPLUG_TIMEOUT) {
+			/* And if anything is actually waiting to be processed */
+			if (hotplug_queue != NULL) {
+				unsigned long long new_min_seqnum;
+
+				/* Select the lowest seqnum in the queue and reset last_hotplug_time_stamp to 0 */
+				new_min_seqnum = G_MAXUINT64;
+				for (i = hotplug_queue; i != NULL; i = g_list_next (i)) {
+					/* this will run since hotplug_queue != NULL */
+					msg = i->data;
+					HAL_INFO (("Looking at %lld", msg->seqnum));
+					if (msg->seqnum < new_min_seqnum)
+						new_min_seqnum = msg->seqnum;
+				}
+				syslog (LOG_ERR, "Timed out waiting for hotplug event %lld. Rebasing to %lld", 
+					last_hotplug_seqnum + 1, new_min_seqnum);
+				HAL_ERROR (("Timed out waiting for hotplug event %lld. Rebasing to %lld", 
+					    last_hotplug_seqnum + 1, new_min_seqnum));
+				last_hotplug_seqnum = new_min_seqnum - 1;
+				last_hotplug_time_stamp = 0;
+
+
+
+				/* process that event now */
+				hald_helper_hotplug_process_queue ();
+			}
+		}
+	}
+}
+
 /** Increment the hotplug semaphore; useful when not wanting to process
  *  hotplug events for a while, like when e.g. adding a hal device
  *  object (which is an asynchronous operation).
@@ -1379,7 +1561,7 @@ hotplug_sem_down (void)
 	if (hotplug_counter < 0) {
 		HAL_ERROR (("****************************************"));
 		HAL_ERROR (("****************************************"));
-		HAL_ERROR (("DANGER WILL ROBISON! hotplug semaphore<0!"));
+		HAL_ERROR (("DANGER WILL ROBINSON! hotplug semaphore<0!"));
 		HAL_ERROR (("****************************************"));
 		HAL_ERROR (("****************************************"));
 		hotplug_counter = 0;
@@ -1390,21 +1572,33 @@ hotplug_sem_down (void)
 		hald_helper_hotplug_process_queue ();
 }
 
+/** This variable is TRUE exactly when we are sleeping on the first hotplug event */
+static gboolean hotplug_sleep_first_event = FALSE;
+
 static gboolean
 hald_helper_first_hotplug_event (gpointer data)
 {
 	GList *i;
 	struct hald_helper_msg *msg;
 
-	/* find the lowest seqnum we should start with */
-	for (i = hotplug_queue; i != NULL; i = g_list_next (i)) {
-		msg = (struct hald_helper_msg *) i->data;
-		if (msg->seqnum < last_hotplug_seqnum)
-			last_hotplug_seqnum = msg->seqnum;
-	}
-	--last_hotplug_seqnum;
+	HAL_INFO (("Slept %dms, now processing events", FIRST_HOTPLUG_SLEEP));
 
-	HAL_INFO (("Starting with SEQNUM=%llu", last_hotplug_seqnum+1));
+	/* First process the queue of events receieved while sleeping on the first hotplug events
+	 * in order to find the last seqnum
+	 *
+	 * This list is sorted.
+	 */
+	for (i = hotplug_queue_first; i != NULL; i = g_list_next (i)) {
+		msg = (struct hald_helper_msg *) i->data;
+		HAL_INFO (("*** msg->seqnum = %lld", msg->seqnum));
+		last_hotplug_seqnum = msg->seqnum;
+		last_hotplug_time_stamp = msg->time_stamp;
+	}
+
+	/* Done sleeping on the first hotplug event */
+	hotplug_sleep_first_event = FALSE;
+
+	HAL_INFO (("Starting with last_hotplug_seqnum=%llu", last_hotplug_seqnum));
 
 	hotplug_sem_down ();
 
@@ -1412,6 +1606,15 @@ hald_helper_first_hotplug_event (gpointer data)
 	return FALSE;
 }
 
+static gint
+hald_helper_msg_compare (gconstpointer pa, gconstpointer pb)
+{
+	const struct hald_helper_msg *a = (const struct hald_helper_msg *) pa;
+	const struct hald_helper_msg *b = (const struct hald_helper_msg *) pb;
+
+	return (gint) (a->seqnum - b->seqnum);
+}
+
 static gboolean
 hald_helper_data (GIOChannel *source, 
 		  GIOCondition condition, 
@@ -1461,6 +1664,20 @@ hald_helper_data (GIOChannel *source,
 		goto out;
 	}
 
+	HAL_INFO (("SEQNUM=%lld, TIMESTAMP=%d", msg.seqnum, msg.time_stamp));
+
+	/* For DEBUG: Drop every 16th hotplug event */
+#if 0
+	if ((msg.seqnum&0x0f) == 0x00) {
+		HAL_INFO (("=========================================="));
+		HAL_INFO (("=========================================="));
+		HAL_INFO (("NOTE NOTE: For debugging, deliberately ignoring hotplug event with SEQNUM=%d", msg.seqnum));
+		HAL_INFO (("=========================================="));
+		HAL_INFO (("=========================================="));
+		goto out;
+	}
+#endif
+
 	switch (msg.type) {
 	case HALD_DEVD:
 		hald_helper_device_name (msg.action, msg.seqnum, g_strdup (msg.subsystem),
@@ -1480,33 +1697,51 @@ hald_helper_data (GIOChannel *source,
 			 * @todo TODO: read SEQNUM from sysfs
 			 */
 
-			HAL_WARNING (("First SEQNUM=%llu; sleeping 2500ms to get a few more events", msg.seqnum));
+			HAL_INFO (("First SEQNUM=%llu; sleeping %dms to get a few more events", 
+				   msg.seqnum, FIRST_HOTPLUG_SLEEP));
 
 			hotplug_sem_up ();
-			g_timeout_add (2500, hald_helper_first_hotplug_event, NULL);
+			g_timeout_add (FIRST_HOTPLUG_SLEEP, hald_helper_first_hotplug_event, NULL);
+			hotplug_sleep_first_event = TRUE;
+
+			hotplug_queue_first = g_list_insert_sorted (hotplug_queue_first, 
+								    g_memdup (&msg, sizeof (struct hald_helper_msg)),
+								    hald_helper_msg_compare);
 
 			/* so we only setup one timer */
 			last_hotplug_seqnum = msg.seqnum;
-		}
+			last_hotplug_time_stamp = msg.time_stamp;
 
-		if (msg.seqnum < last_hotplug_seqnum) {
-			/* yikes, this means were started during a hotplug */
-			HAL_WARNING (("Got SEQNUM=%d, but last_hotplug_seqnum=%llu", msg.seqnum, last_hotplug_seqnum));
-
-			/* have to process immediately other we may deadlock due to
-			 * the hotplug semaphore */
-			hald_helper_hotplug (msg.action, msg.seqnum, g_strdup (msg.subsystem), 
-					     g_strdup (msg.sysfs_path), &msg);
-			/* still need to process the queue though */
-			hald_helper_hotplug_process_queue ();
 			goto out;
 		}
 
-		/* Queue up this hotplug event and process the queue */
-		HAL_INFO (("Queing up seqnum=%llu, sysfspath=%s, subsys=%s", 
-			   msg.seqnum, msg.sysfs_path, msg.subsystem));
-		hotplug_queue = g_list_append (hotplug_queue, g_memdup (&msg, sizeof (struct hald_helper_msg)));
-		hald_helper_hotplug_process_queue ();
+		/* If sleeping on the first hotplug event queue up the events in a special queue */
+		if (hotplug_sleep_first_event) {
+			HAL_INFO (("first hotplug sleep; got SEQNUM=%d", msg.seqnum));
+			hotplug_queue_first = g_list_insert_sorted (hotplug_queue_first, 
+								    g_memdup (&msg, sizeof (struct hald_helper_msg)),
+								    hald_helper_msg_compare);
+		} else {
+
+			if (msg.seqnum < last_hotplug_seqnum) {
+				/* yikes, this means were started during a hotplug */
+				HAL_WARNING (("Got SEQNUM=%llu, but last_hotplug_seqnum=%llu", 
+					      msg.seqnum, last_hotplug_seqnum));
+                                /* have to process immediately other we may deadlock due to the hotplug semaphore */
+				hald_helper_hotplug (msg.action, msg.seqnum, g_strdup (msg.subsystem), 	 
+						     g_strdup (msg.sysfs_path), &msg); 	 
+				/* still process the queue though */
+				hald_helper_hotplug_process_queue ();
+
+			} else {
+				/* Queue up this hotplug event and process the queue */
+				HAL_INFO (("Queing up seqnum=%llu, sysfspath=%s, subsys=%s", 
+					   msg.seqnum, msg.sysfs_path, msg.subsystem));
+				hotplug_queue = g_list_append (hotplug_queue, 
+							       g_memdup (&msg, sizeof (struct hald_helper_msg)));
+				hald_helper_hotplug_process_queue ();
+			}
+		}
 		break;
 	}
 out:
