@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: usb_bus_device.c,v 1.13 2004/09/19 14:47:28 david Exp $
+ * CVSID: $Id: usb_bus_device.c,v 1.15.2.1 2005/01/24 21:28:33 david Exp $
  *
  * USB bus device
  *
@@ -237,6 +237,7 @@ usb_ids_load (const char *path)
 		printf
 		    ("Couldn't allocate %d bytes for USB database file\n",
 		     usb_ids_len);
+		fclose(fp);
 		return FALSE;
 	}
 
@@ -245,9 +246,11 @@ usb_ids_load (const char *path)
 		printf ("Error loading USB database file\n");
 		free (usb_ids);
 		usb_ids = NULL;
+		fclose(fp);
 		return FALSE;
 	}
 
+	fclose(fp);
 	return TRUE;
 }
 
@@ -704,8 +707,7 @@ usb_merge_info_from_proc (HalDevice* d)
 
 			/* This is quite a hack */
 			port_number = 0;
-			for (i = len - 1; i > 0 && isdigit (bus_id[i]);
-			     --i) {
+			for (i = len - 1; i > 0 && isdigit (bus_id[i]); --i) {
 				digit = (int) (bus_id[i] - '0');
 				port_number *= 10;
 				port_number += digit;
@@ -810,7 +812,6 @@ usb_device_pre_process (BusDeviceHandler *self,
 	char *product_name_kernel = NULL;
 	char numeric_name[32];
 	struct sysfs_attribute *cur;
-	int len;
 
 	dlist_for_each_data (sysfs_get_device_attributes (device), cur,
 			     struct sysfs_attribute) {
@@ -820,9 +821,11 @@ usb_device_pre_process (BusDeviceHandler *self,
 					      SYSFS_NAME_LEN) != 0)
 			continue;
 
+		if (cur->len <= 0)
+			continue;
+
 		/* strip whitespace */
-		len = strlen (cur->value);
-		for (i = len - 1; i >= 0 && isspace (cur->value[i]); --i)
+		for (i = cur->len - 1; i >= 0 && isspace (cur->value[i]); --i)
 			cur->value[i] = '\0';
 
 		/*printf("attr_name=%s -> '%s'\n", attr_name, cur->value); */
@@ -837,8 +840,7 @@ usb_device_pre_process (BusDeviceHandler *self,
 		else if (strcmp (attr_name, "bMaxPower") == 0)
 			hal_device_property_set_int (d, "usb_device.max_power",
 					     parse_dec (cur->value));
-		else if (strcmp (attr_name, "serial") == 0
-			 && strlen (cur->value) > 0)
+		else if (strcmp (attr_name, "serial") == 0 && cur->len > 0)
 			hal_device_property_set_string (d, "usb_device.serial",
 						cur->value);
 		else if (strcmp (attr_name, "bmAttributes") == 0) {
@@ -944,7 +946,7 @@ BusDeviceHandler usb_bus_handler = {
 	usb_device_shutdown,       /**< shutdown function */
 	bus_device_tick,           /**< timer function */
 	usb_device_accept,         /**< accept function */
- 	bus_device_visit,          /**< visitor function */
+	bus_device_visit,          /**< visitor function */
 	bus_device_removed,        /**< device is removed */
 	usb_device_compute_udi,    /**< UDI computing function */
 	usb_device_pre_process,    /**< add more properties */
