@@ -1,5 +1,5 @@
 /***************************************************************************
- * CVSID: $Id: libhal.c,v 1.26 2004/09/17 17:05:48 joe Exp $
+ * CVSID: $Id: libhal.c,v 1.32.2.1 2005/01/07 03:18:53 david Exp $
  *
  * libhal.c : HAL daemon C convenience library
  *
@@ -34,6 +34,25 @@
 
 #include "libhal.h"
 
+#ifdef ENABLE_NLS
+# include <libintl.h>
+# define _(String) dgettext (GETTEXT_PACKAGE, String)
+# ifdef gettext_noop
+#   define N_(String) gettext_noop (String)
+# else
+#   define N_(String) (String)
+# endif
+#else
+/* Stubs that do something close enough.  */
+# define textdomain(String) (String)
+# define gettext(String) (String)
+# define dgettext(Domain,Message) (Message)
+# define dcgettext(Domain,Message,Type) (Message)
+# define bindtextdomain(Domain,Directory) (Domain)
+# define _(String)
+# define N_(String) (String)
+#endif
+
 /**
  * @defgroup LibHal HAL convenience library
  * @brief A convenience library used to communicate with the HAL daemon
@@ -104,6 +123,7 @@ struct LibHalProperty_s {
 struct LibHalContext_s {
 	DBusConnection *connection;           /**< D-BUS connection */
 	dbus_bool_t is_initialized;           /**< Are we initialised */
+	dbus_bool_t is_shutdown;              /**< Have we been shutdown */
 	dbus_bool_t cache_enabled;            /**< Is the cache enabled */
 	const LibHalFunctions *functions;     /**< Callback functions */
 	void *user_data;                      /**< User data */
@@ -293,6 +313,9 @@ hal_free_property_set (LibHalPropertySet * set)
 	LibHalProperty *p;
 	LibHalProperty *q;
 
+	if (set == NULL)
+		return;
+
 	for (p = set->properties_head; p != NULL; p = q) {
 		free (p->key);
 		if (p->type == DBUS_TYPE_STRING)
@@ -432,6 +455,9 @@ filter_func (DBusConnection * connection,
 	DBusError error;
 	LibHalContext *ctx = (LibHalContext *) user_data;
 
+	if (ctx->is_shutdown)
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
 	dbus_error_init (&error);
 
 	object_path = dbus_message_get_path (message);
@@ -449,7 +475,7 @@ filter_func (DBusConnection * connection,
 			}
 			dbus_free (udi);
 		}
-		return DBUS_HANDLER_RESULT_HANDLED;
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	} else
 	    if (dbus_message_is_signal
 		(message, "org.freedesktop.Hal.Manager",
@@ -463,7 +489,7 @@ filter_func (DBusConnection * connection,
 			}
 			dbus_free (udi);
 		}
-		return DBUS_HANDLER_RESULT_HANDLED;
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	} else
 	    if (dbus_message_is_signal
 		(message, "org.freedesktop.Hal.Manager",
@@ -482,7 +508,7 @@ filter_func (DBusConnection * connection,
 			dbus_free (udi);
 			dbus_free (capability);
 		}
-		return DBUS_HANDLER_RESULT_HANDLED;
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	} else
 	    if (dbus_message_is_signal
 		(message, "org.freedesktop.Hal.Device", "Condition")) {
@@ -501,7 +527,7 @@ filter_func (DBusConnection * connection,
 
 			dbus_free (condition_name);
 		}
-		return DBUS_HANDLER_RESULT_HANDLED;
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	} else
 	    if (dbus_message_is_signal
 		(message, "org.freedesktop.Hal.Device",
@@ -540,7 +566,7 @@ filter_func (DBusConnection * connection,
 			}
 
 		}
-		return DBUS_HANDLER_RESULT_HANDLED;
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
 
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -557,6 +583,9 @@ static LibHalFunctions hal_null_functions = {
 };
 #endif /* DOXYGEN_SHOULD_SKIP_THIS */
 
+/* for i18n purposes */
+static dbus_bool_t hal_already_initialized_once = FALSE;
+
 /** Initialize the HAL library. 
  *
  *  @param  cb_functions	  Callback functions. If this is set top NULL
@@ -581,6 +610,13 @@ hal_initialize (const LibHalFunctions * cb_functions,
 {
 	DBusError error;
 	LibHalContext *ctx;
+
+	if (!hal_already_initialized_once) {
+		bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
+		bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+		
+		hal_already_initialized_once = TRUE;
+	}
 	
 	ctx = malloc (sizeof (LibHalContext));
 	if (ctx == NULL) {
@@ -589,6 +625,9 @@ hal_initialize (const LibHalFunctions * cb_functions,
 		return NULL;
 	}
 
+	ctx->is_initialized = FALSE;
+	ctx->is_shutdown = FALSE;
+
 	ctx->cache_enabled = use_cache;
 
 	ctx->functions = cb_functions;
@@ -635,7 +674,6 @@ hal_initialize (const LibHalFunctions * cb_functions,
 	}
 
 	ctx->is_initialized = TRUE;
-
 	return ctx;
 }
 
@@ -648,11 +686,42 @@ hal_initialize (const LibHalFunctions * cb_functions,
 int
 hal_shutdown (LibHalContext *ctx)
 {
+	DBusError error;
+
 	if (!ctx->is_initialized)
 		return 1;
 
-	/** @todo cleanup */
-	free (ctx);
+	/* unsubscribe the match rule we added in initialize; this is safe even with multiple
+	 * instances of libhal running - see the dbus docs */
+	dbus_error_init (&error);
+	dbus_bus_remove_match (ctx->connection,
+			       "type='signal',"
+			       "interface='org.freedesktop.Hal.Manager',"
+			       "sender='org.freedesktop.Hal',"
+			       "path='/org/freedesktop/Hal/Manager'", &error);
+	if (dbus_error_is_set (&error)) {
+		fprintf (stderr, "%s %d : Error removing match rule, error=%s\r\n",
+			 __FILE__, __LINE__, error.message);
+	}
+
+	/* TODO: remove all other match rules */
+
+	/* set a flag so we don't propagte callbacks from this context anymore */
+	ctx->is_shutdown = TRUE;
+
+	/* yikes, it's dangerous to unref the connection since it will terminate the process
+	 * because this connection may be shared so we cannot set the exit_on_disconnect flag
+	 *
+	 * so we don't do that right now 
+	 *
+	 */
+	/*dbus_connection_unref (ctx->connection);*/
+
+	/* we also refuse to free the resources as filter_function may reference these 
+	 * 
+	 * should free async when our connection goes away.
+	 */
+	/* free (ctx); */
 	return 0;
 }
 
@@ -676,6 +745,8 @@ hal_get_all_devices (LibHalContext *ctx, int *num_devices)
 	char **device_names;
 	char **hal_device_names;
 
+	*num_devices = 0;
+
 	message = dbus_message_new_method_call ("org.freedesktop.Hal",
 						"/org/freedesktop/Hal/Manager",
 						"org.freedesktop.Hal.Manager",
@@ -2159,15 +2230,13 @@ dbus_bool_t
 hal_device_query_capability (LibHalContext *ctx, 
 			     const char *udi, const char *capability)
 {
+	dbus_bool_t ret;
 	char *caps;
 
 	caps = hal_device_get_property_string (ctx, udi, "info.capabilities");
-
-	if (caps != NULL)
-		if (strstr (caps, capability) != NULL)
-			return TRUE;
-
-	return FALSE;
+	ret = (caps != NULL && strstr (caps, capability) != NULL);
+	hal_free_string (caps);
+	return ret;
 }
 
 /** Find devices with a given capability. 
